# entities/base.py
from __future__ import annotations

from dataclasses import dataclass, field
from typing import (
    Any,
    ClassVar,
    Dict,
    List,
    Optional,
    Set,
    TypeVar,
    TYPE_CHECKING,
)

TEntity = TypeVar("TEntity", bound="BaseEntity")

if TYPE_CHECKING:
    from ..client import DataClient


from typing import Generic, Callable, TypeVar, Optional

T = TypeVar("T")


class Field(Generic[T]):
    """
    Descriptor mapping an attribute to a key in `entity.data`.

    Example:
        title: str = Field("title", default="")
    """

    def __init__(
        self,
        key: Optional[str] = None,
        *,
        default: Optional[T] = None,
        default_factory: Optional[Callable[[], T]] = None,
        read_only: bool = False,
    ) -> None:
        self.key = key
        self.default = default
        self.default_factory = default_factory
        self.read_only = read_only
        self.name: Optional[str] = None

    def __set_name__(self, owner, name: str) -> None:
        if self.key is None:
            self.key = name
        self.name = name

    def __get__(self, instance, owner=None) -> T:
        if instance is None:
            return self

        key = self.key
        assert key is not None

        # Auto-fetch lazy entities when accessing a field other than 'urn'
        if key != "urn" and key not in instance.data:
            # Check if this is a lazy entity (only has 'urn' in data)
            if len(instance.data) == 1 and "urn" in instance.data:
                instance.refresh()

        if key in instance.data:
            return instance.data[key]

        if self.default_factory is not None:
            value = self.default_factory()
            instance.data[key] = value
            return value

        return self.default

    def __set__(self, instance, value: T) -> None:
        if self.read_only:
            raise AttributeError(f"Field '{self.name}' is read-only")

        key = self.key
        assert key is not None

        instance.data[key] = value

        # Track as dirty if supported
        dirty = getattr(instance, "_dirty_fields", None)
        if isinstance(dirty, set):
            dirty.add(key)

        # Auto-sync on write if enabled
        if getattr(instance, "sync", False):
            instance.save(only_dirty=True)


@dataclass
class BaseEntity:
    """
    Lightweight proxy for a single API entity.

    Subclasses must define:
      - ENDPOINT   (e.g. "articles")
      - URN_PREFIX (e.g. "urn:article:")
    """

    client: "DataClient"
    data: Dict[str, Any] = field(default_factory=dict)

    sync: bool = field(default=True, repr=False, compare=False)

    _dirty_fields: Set[str] = field(default_factory=set, repr=False, compare=False)

    ENDPOINT: ClassVar[str] = ""
    URN_PREFIX: ClassVar[str] = ""

    # ------------------------------------------------------------------ #
    # URN handling
    # ------------------------------------------------------------------ #

    @classmethod
    def normalize_urn(cls, urn_or_slug: str) -> str:
        urn_or_slug = urn_or_slug.lstrip("/")

        # If a full URN like "urn:article:slug" (contains ':') is provided,
        # return the bare slug (the part after the last colon).
        if ":" in urn_or_slug:
            return urn_or_slug.rsplit(":", 1)[1]

        # Otherwise, if the class has a URN_PREFIX and the input starts with it,
        # strip the prefix.
        if cls.URN_PREFIX and urn_or_slug.startswith(cls.URN_PREFIX):
            return urn_or_slug[len(cls.URN_PREFIX) :]

        return urn_or_slug

    @property
    def urn(self) -> str:
        """URN of the underlying entity."""
        return self.data["urn"]

    @staticmethod
    def _extract_result(payload: Any) -> Any:
        """
        Unwrap the API response and return the actual entity payload.

        The Wisefood API usually nests data under a "result" key.
        """
        if isinstance(payload, dict) and "result" in payload:
            return payload["result"]
        return payload

    # ------------------------------------------------------------------ #
    # CRUD (class methods)
    # ------------------------------------------------------------------ #

    @classmethod
    def get(cls, client: "DataClient", urn: str) -> "BaseEntity":
        """Fetch a single entity by URN (or slug) and return a proxy."""
        full = cls.normalize_urn(urn)
        resp = client.get(f"{cls.ENDPOINT}/{full}")
        payload = resp.json()
        result = cls._extract_result(payload)
        return cls(client=client, data=result)

    @classmethod
    def create(cls, client: "DataClient", *, urn: str, **fields: Any) -> "BaseEntity":
        """Create a new entity and return a proxy for it."""
        payload = {"urn": cls.normalize_urn(urn), **fields}
        resp = client.post(cls.ENDPOINT, json=payload)
        payload = resp.json()
        result = cls._extract_result(payload)
        return cls(client=client, data=result)

    @classmethod
    def enhance(
        cls, client: "DataClient", *, urn: str, agent: str, **fields: Any
    ) -> "BaseEntity":
        """Enhance an existing entity and return a proxy for it."""
        full_urn = cls.normalize_urn(urn)
        payload = {"agent": agent, **fields}
        resp = client.post(f"{cls.ENDPOINT}/{full_urn}/enhance", json=payload)
        payload = resp.json()
        result = cls._extract_result(payload)
        return cls(client=client, data=result)

    # ------------------------------------------------------------------ #
    # CRUD (instance methods)
    # ------------------------------------------------------------------ #

    def refresh(self) -> None:
        """Reload the entity data from the API."""
        resp = self.client.get(f"{self.ENDPOINT}/{self.urn}")
        payload = resp.json()
        self.data = self._extract_result(payload)

    def save(self, *, only_dirty: bool = False) -> None:
        """
        Persist local changes to the API using PATCH.

        If only_dirty=True, only fields that have been changed via Field
        descriptors are sent (based on `_dirty_fields`).
        """
        if only_dirty and self._dirty_fields:
            keys = self._dirty_fields
            body = {
                k: self.data[k]
                for k in keys
                if k not in {"id", "creator", "created_at", "updated_at"}
            }
        else:
            body = {
                k: v
                for k, v in self.data.items()
                if k not in {"id", "creator", "created_at", "updated_at"}
            }

        if not body:
            return  # nothing to send

        resp = self.client.patch(f"{self.ENDPOINT}/{self.urn}", json=body)
        payload = resp.json()
        self.data = self._extract_result(payload)
        self._dirty_fields.clear()

    def delete(self) -> None:
        self.client.delete(f"{self.ENDPOINT}/{self.urn}")

    def enhance_self(self, *, agent: str, **fields: Any) -> None:
        """
        Enhance this entity using an AI agent and update its data.

        Args:
            agent: The AI agent identifier to use for enhancement
            **fields: Additional fields to send with the enhancement request
        """
        payload = {"agent": agent, **fields}
        resp = self.client.patch(f"{self.ENDPOINT}/{self.urn}/enhance", json=payload)
        payload = resp.json()
        self.data = self._extract_result(payload)
        self._dirty_fields.clear()

        return self.get(self.client, self.urn)

    # ------------------------------------------------------------------ #
    # Representation / display helpers
    # ------------------------------------------------------------------ #

    def __repr__(self) -> str:
        """Short, machine-oriented representation."""
        return f"<{self.__class__.__name__} urn='{self.urn}'>"

    def __str__(self) -> str:
        """Compact human-oriented summary."""
        title = self.data.get("title")
        if title:
            return f"{self.__class__.__name__}(urn='{self.urn}', title='{title}')"
        return f"{self.__class__.__name__}(urn='{self.urn}')"

    def json(self) -> None:
        """Pretty-print the full metadata payload."""
        import json

        print(json.dumps(self.data, indent=4, ensure_ascii=False))

    def dict(self) -> Dict[str, Any]:
        """Return the full metadata payload as a dictionary."""
        return self.data

    def show(self) -> None:
        """Pretty-print the full metadata payload via Pandas"""
        import pandas as pd

        df = pd.json_normalize(self.data)
        pd.set_option("display.max_columns", None)
        pd.set_option("display.width", None)
        print(df.T)


class BaseCollectionProxy:
    """
    Indexable / sliceable view over a collection of entities.

    The list endpoint is expected to return either:
      - {"result": ["urn:...","urn:...", ...]}
      - {"result": [{"urn": "...", ...}, ...]}
    """

    ENTITY_CLS: ClassVar[type[BaseEntity]] = BaseEntity
    ENDPOINT: ClassVar[str] = ""
    DEFAULT_PAGE_SIZE: ClassVar[int] = 100  # used for index/completions

    def __init__(self, client: "DataClient") -> None:
        self.client = client
        self._urns: Optional[List[str]] = None

    # ------------------------------------------------------------------ #
    # Low-level fetching helpers
    # ------------------------------------------------------------------ #

    def _parse_list_result(self, payload: Any) -> List[str]:
        """Normalize list responses into a list of URNs."""
        result = payload.get("result", payload)

        if isinstance(result, list) and (not result or isinstance(result[0], str)):
            return result
        if isinstance(result, list) and isinstance(result[0], dict):
            return [item["urn"] for item in result]

        raise ValueError(f"Unexpected list endpoint format: {result!r}")

    def _fetch_urns(self, *, limit: int, offset: int = 0) -> List[str]:
        """Fetch a page of URNs using limit/offset."""
        resp = self.client.get(self.ENDPOINT, limit=limit, offset=offset)
        payload = resp.json()
        return self._parse_list_result(payload)

    def _ensure_index(self) -> None:
        """Populate an initial page of URNs for len()/iteration/completions."""
        if self._urns is not None:
            return
        self._urns = self._fetch_urns(
            limit=self.DEFAULT_PAGE_SIZE,
            offset=0,
        )

    def _get_entity(self, urn: str, *, lazy: bool = False) -> BaseEntity:
        """
        Return an entity proxy for a given URN.

        Args:
            urn: The URN or slug of the entity
            lazy: If True, return a lazy proxy without fetching data immediately.
                  The entity will only contain the URN until accessed.
        """
        urn = urn.lstrip("/")

        if lazy:
            # Create a lazy proxy with just the URN, no API call
            full_urn = (
                urn
                if urn.startswith(self.ENTITY_CLS.URN_PREFIX)
                else self.ENTITY_CLS.URN_PREFIX + urn
            )
            return self.ENTITY_CLS(client=self.client, data={"urn": full_urn})

        return self.ENTITY_CLS.get(self.client, urn)

    # ------------------------------------------------------------------ #
    # Creation helpers
    # ------------------------------------------------------------------ #

    def create(self, *, urn: str, **fields: Any) -> BaseEntity:
        """
        Create a new entity through the proxy and return its proxy object.

        Keeps the cached index in sync when it has already been populated.
        """
        entity = self.ENTITY_CLS.create(self.client, urn=urn, **fields)

        if self._urns is not None:
            full_urn = entity.urn
            if full_urn not in self._urns:
                self._urns.append(full_urn)

        return entity

    # ------------------------------------------------------------------ #
    # Enhancement helpers
    # ------------------------------------------------------------------ #
    def enhance(self, agent: str, *, urn: str, **fields: Any) -> BaseEntity:
        """
        Enhance an existing entity through the proxy and return its proxy object.

        Keeps the cached index in sync when it has already been populated.
        """
        entity = self.ENTITY_CLS.enhance(self.client, urn=urn, agent=agent, **fields)

        if self._urns is not None:
            full_urn = entity.urn
            if full_urn not in self._urns:
                self._urns.append(full_urn)

        return entity

    # ------------------------------------------------------------------ #
    # Search helpers
    # ------------------------------------------------------------------ #
    def search(
        self,
        q: str,
        fl: Optional[List[str]] = None,
        limit: int = 10,
        offset: int = 0,
        fq: Optional[List[str]] = None,
        sort: Optional[str] = None,
        fields: Optional[List[str]] = None,
        facet_limit: int = 50,
        highlight: bool = False,
        highlight_fields: Optional[List[str]] = None,
        highlight_pre_tag: str = "<em>",
        highlight_post_tag: str = "</em>",
    ) -> List[BaseEntity]:
        """Search entities with optional filters and highlighting."""
        payload = {
            "q": q,
            "limit": limit,
            "offset": offset,
        }
        if fl is not None:
            payload["fl"] = fl
        if fq is not None:
            payload["fq"] = fq
        if sort is not None:
            payload["sort"] = sort
        if fields is not None:
            payload["fields"] = fields
        if facet_limit != 50:
            payload["facet_limit"] = facet_limit
        if highlight:
            payload["highlight"] = highlight
            if highlight_fields is not None:
                payload["highlight_fields"] = highlight_fields
            payload["highlight_pre_tag"] = highlight_pre_tag
            payload["highlight_post_tag"] = highlight_post_tag

        resp = self.client.post(f"{self.ENDPOINT}/search", json=payload).json()
        results = resp.get("result", {}).get("results", [])

        # Parse results into entities
        entities = []
        for item in results:
            if isinstance(item, dict) and "urn" in item:
                entity = self.ENTITY_CLS(client=self.client, data=item)
                entities.append(entity)
            elif isinstance(item, str):
                entities.append(self._get_entity(item))

        return entities

    # ------------------------------------------------------------------ #
    # Python container protocol
    # ------------------------------------------------------------------ #

    def __len__(self) -> int:
        self._ensure_index()
        return len(self._urns or [])

    def __iter__(self):
        """Iterate over entity proxies (one API call per entity)."""
        self._ensure_index()
        for urn in self._urns or []:
            yield self._get_entity(urn)

    def __getitem__(self, key):
        """
        Support:
          - proxy[0]           → entity by position
          - proxy[1:10]        → list of entities (using limit/offset)
          - proxy["slug"]      → entity by slug
          - proxy["urn:..."]   → entity by full URN
          - proxy["text"]      → fuzzy URN search (substring)
        """

        # integer index → use cached index
        if isinstance(key, int):
            self._ensure_index()
            if self._urns is None:
                raise IndexError("Index not loaded")
            urn = self._urns[key]
            return self._get_entity(urn)

        # slice → use limit/offset when possible
        if isinstance(key, slice):
            start = key.start or 0
            stop = key.stop
            step = key.step or 1

            if step != 1:
                raise ValueError("Step other than 1 is not supported for slices.")

            if stop is None:
                raise ValueError("Open-ended slices are not supported; specify stop.")

            limit = max(0, stop - start)
            if limit == 0:
                return []

            urns = self._fetch_urns(limit=limit, offset=start)
            # Return lazy proxies - entities are only fetched when actually accessed
            return [self._get_entity(u, lazy=True) for u in urns]

        # string: URN / slug / fuzzy search
        if isinstance(key, str):
            key = key.lstrip("/")
            self._ensure_index()

            prefix = self.ENTITY_CLS.URN_PREFIX

            # full URN - check cache first, then fetch from API
            if key.startswith(prefix):
                if self._urns is not None and key in self._urns:
                    return self._get_entity(key)
                # Not in cache, try fetching from API
                return self._get_entity(key)

            # slug → full URN - check cache first, then fetch from API
            full = prefix + key
            if self._urns is not None and full in self._urns:
                return self._get_entity(full)

            # Not in cache, try fetching from API directly
            # This handles the case where entity exists on server but not in local index
            return self._get_entity(key)

        raise TypeError(f"Unsupported key type: {type(key)!r}")

    # ------------------------------------------------------------------ #
    # Helpers for editor / IPython completion
    # ------------------------------------------------------------------ #

    def slugs(self) -> List[str]:
        """Return slug forms of all known URNs (URN prefix stripped)."""
        self._ensure_index()
        prefix = self.ENTITY_CLS.URN_PREFIX
        return [u.replace(prefix, "") for u in (self._urns or [])]

    def __dir__(self):
        slugs = self.slugs()
        return list(super().__dir__()) + slugs

    def _ipython_key_completions_(self):
        try:
            return self.slugs()
        except Exception:
            return []



"""wisefood - A small client for accessing and populating the data infrastructure of the WiseFood platform."""

from .client import DataClient, Credentials
from .api_client import Client

__all__ = ["Client", "DataClient", "Credentials"]
__version__ = "0.0.1"



from wisefood.exceptions import DataError, error_from_response


class StubResponse:
    def __init__(self, status_code, payload):
        self.status_code = status_code
        self._payload = payload

    def json(self):
        return self._payload

    @property
    def text(self):
        return ""


def test_validation_error_includes_detail_message():
    payload = {
        "detail": [
            {"loc": ["body", "title"], "msg": "field required", "type": "value_error.missing"}
        ]
    }
    resp = StubResponse(422, payload)

    err = error_from_response(resp)

    assert isinstance(err, DataError)
    assert "body.title: field required" in str(err)
    assert err.errors == payload["detail"]
    assert err.response_body == payload


def test_enveloped_validation_error_includes_detail_message():
    payload = {
        "success": False,
        "error": {
            "title": "RequestValidationError",
            "detail": "Validation failed",
            "code": "request/unprocessable",
            "errors": [
                {
                    "type": "string_pattern_mismatch",
                    "loc": ["body", "urn"],
                    "msg": "String should match pattern '^[a-z0-9]+(?:[-_][a-z0-9]+)*$'",
                    "input": "urn:article:bad",
                    "ctx": {"pattern": "^[a-z0-9]+(?:[-_][a-z0-9]+)*$"},
                }
            ],
        },
        "help": "http://example/api/v1/articles",
    }
    resp = StubResponse(422, payload)

    err = error_from_response(resp)

    assert isinstance(err, DataError)
    assert "Validation failed: body.urn: String should match pattern" in str(err)
    assert err.errors == payload["error"]["errors"]



from __future__ import annotations

from dataclasses import dataclass
from typing import Any, Dict, Optional


@dataclass
class APIError(Exception):
    """
    Base client-side API error.

    Mirrors the server-side structure:
        {
          "success": false,
          "error": {
              "title": "...",
              "detail": "...",
              "code": "resource/not_found",
              ...
          },
          "help": "http://.../api/v1/articles/hello"
        }
    """

    status_code: int
    detail: str = ""
    code: Optional[str] = None          # e.g. "resource/not_found"
    title: Optional[str] = None         # e.g. "NotFoundError"
    errors: Any = None                  # nested errors (validation etc.)
    extra: Dict[str, Any] = None        # extra info from server
    help_url: Optional[str] = None      # value of top-level 'help' field
    response_body: Any = None           # raw parsed JSON of the response

    def __post_init__(self) -> None:
        if self.extra is None:
            self.extra = {}
        # Default message for Exception.__str__
        msg = self.detail or self.title or f"HTTP {self.status_code}"
        super().__init__(msg)

    @property
    def retryable(self) -> bool:
        """Whether a retry might make sense (for client backoff logic)."""
        return (
            self.status_code in (429, 503, 504)
            or 500 <= self.status_code < 600
        )


# -------------------------------------------------
# Typed client-side exceptions
# -------------------------------------------------

class InvalidError(APIError):
    pass


class DataError(APIError):
    pass


class AuthenticationError(APIError):
    pass


class AuthorizationError(APIError):
    pass


class NotFoundError(APIError):
    pass


class NotAllowedError(APIError):
    pass


class ConflictError(APIError):
    pass


class RateLimitError(APIError):
    pass


class InternalError(APIError):
    pass


class BadGatewayError(APIError):
    pass


class ServiceUnavailableError(APIError):
    pass


class GatewayTimeoutError(APIError):
    pass


# -------------------------------------------------
# Mapping helpers
# -------------------------------------------------

# Map server-side `code` → specific client exception
_CODE_TO_EXCEPTION = {
    "request/invalid": InvalidError,
    "request/unprocessable": DataError,
    "auth/unauthorized": AuthenticationError,
    "auth/forbidden": AuthorizationError,
    "resource/not_found": NotFoundError,
    "request/not_allowed": NotAllowedError,
    "resource/conflict": ConflictError,
    "quota/rate_limited": RateLimitError,
    "server/internal": InternalError,
    "upstream/bad_gateway": BadGatewayError,
    "upstream/unavailable": ServiceUnavailableError,
    "upstream/timeout": GatewayTimeoutError,
}

# Fallback mapping by HTTP status code
_STATUS_TO_EXCEPTION = {
    400: InvalidError,
    401: AuthenticationError,
    403: AuthorizationError,
    404: NotFoundError,
    405: NotAllowedError,
    409: ConflictError,
    422: DataError,
    429: RateLimitError,
    500: InternalError,
    502: BadGatewayError,
    503: ServiceUnavailableError,
    504: GatewayTimeoutError,
}


def _pick_exception_class(status_code: int, code: Optional[str]) -> type[APIError]:
    if code and code in _CODE_TO_EXCEPTION:
        return _CODE_TO_EXCEPTION[code]
    if status_code in _STATUS_TO_EXCEPTION:
        return _STATUS_TO_EXCEPTION[status_code]
    return APIError


def _format_detail(detail: Any) -> str:
    """
    Turn common validation-detail shapes into a readable string.

    FastAPI-style validation errors often look like:
        {"detail": [{"loc": ["body", "field"], "msg": "field required", ...}, ...]}
    """
    if isinstance(detail, str):
        return detail

    if isinstance(detail, list):
        parts = []
        for item in detail:
            if isinstance(item, dict):
                loc = item.get("loc") or []
                loc_str = ".".join(str(p) for p in loc) if loc else ""
                msg = item.get("msg") or item.get("message") or str(item)
                parts.append(f"{loc_str}: {msg}" if loc_str else str(msg))
            else:
                parts.append(str(item))
        return "; ".join(parts)

    return str(detail)


def error_from_response(response) -> APIError:
    """
    Build a concrete APIError subclass from a `requests.Response`.

    Expected Wisefood error payload shape:
        {
          "success": false,
          "error": {
            "title": "NotFoundError",
            "detail": "Article with URN urn:article:hello not found.",
            "code": "resource/not_found",
            ... maybe other keys ...
          },
          "help": "http://.../api/v1/articles/hello"
        }

    If the body is not JSON or doesn't match the envelope, we still
    build a generic APIError with whatever information we can.
    """

    status_code = response.status_code

    body: Any
    try:
        body = response.json()
    except Exception:
        # Non-JSON error
        return APIError(
            status_code=status_code,
            detail=response.text or f"HTTP {status_code}",
            response_body=None,
        )

    success = body.get("success")
    error_block = body.get("error") if isinstance(body, dict) else None
    help_url = body.get("help") if isinstance(body, dict) else None

    # If the envelope is not there, fall back to generic
    if success is not False or not isinstance(error_block, dict):
        # Special-case validation errors (422) that might not use the envelope
        if status_code == 422 and isinstance(body, dict) and "detail" in body:
            detail = _format_detail(body.get("detail", ""))
            return DataError(
                status_code=status_code,
                detail=detail,
                errors=body.get("detail"),
                extra={k: v for k, v in body.items() if k != "detail"},
                response_body=body,
                help_url=help_url,
            )

        return APIError(
            status_code=status_code,
            detail=str(body),
            response_body=body,
            help_url=help_url,
        )

    # Extract structured fields from the error block
    title = error_block.get("title")
    detail = error_block.get("detail", "")
    code = error_block.get("code")
    errors = error_block.get("errors")  # optional

    # Surface validation error details in the exception message
    formatted_errors = _format_detail(errors) if errors else ""
    if status_code == 422 and formatted_errors:
        detail = f"{detail}: {formatted_errors}" if detail else formatted_errors

    extra = {
        k: v
        for k, v in error_block.items()
        if k not in {"title", "detail", "code", "errors"}
    }

    exc_cls = _pick_exception_class(status_code, code)

    return exc_cls(
        status_code=status_code,
        detail=detail,
        code=code,
        title=title,
        errors=errors,
        extra=extra,
        help_url=help_url,
        response_body=body,
    )


def raise_for_api_error(response) -> None:
    """
    Inspect a `requests.Response` and raise a suitable APIError subclass
    if the Wisefood API indicates failure.

    Usage in your client:

        resp = self.get(\"articles/hello\")
        raise_for_api_error(resp)
        data = resp.json()[\"result\"]

    Behavior:
    - If HTTP status is 2xx and `success` is True → returns silently.
    - If HTTP status >= 400 or `success` is False → raises APIError subclass.
    """
    status = response.status_code

    # 2xx → might still be envelope-based error, so we check body
    try:
        body = response.json()
    except Exception:
        if status >= 400:
            raise error_from_response(response)
        return

    if isinstance(body, dict) and "success" in body:
        if body.get("success") is True and status < 400:
            return
        # success is False or HTTP error
        raise error_from_response(response)

    # No `success` field; rely on HTTP status
    if status >= 400:
        raise error_from_response(response)



"""
IPython/Jupyter key-completion support for bracket-style access:

    client.articles["he<TAB>

This integrates with IPython’s tab completion protocol and works for
any collection subclassing BaseCollectionProxy.
"""

import re
from IPython import get_ipython
from wisefood.entities.base import BaseCollectionProxy


def completion_for_collections(self, event):
    """
    Return suggested slugs for expressions of the form:

        <object>.<collection>["<prefix>

    Only triggers when <collection> is a BaseCollectionProxy.
    """

    line = event.line

    # Match:   variable.collection["prefix
    match = re.search(r'(\w+)\.(\w+)\["([^"]*)$', line)
    if not match:
        return []

    var_name, attr_name, prefix = match.groups()

    shell = get_ipython()
    if shell is None:
        return []

    # Resolve the base object (e.g. "client")
    base_obj = shell.user_ns.get(var_name)
    if base_obj is None:
        return []

    # Resolve the collection proxy (e.g. "articles")
    collection = getattr(base_obj, attr_name, None)
    if not isinstance(collection, BaseCollectionProxy):
        return []

    try:
        slugs = collection.slugs()
    except Exception:
        return []

    # Prefix-based completion (not fuzzy)
    return [s for s in slugs if s.startswith(prefix)]


# Register the completer with IPython
ip = get_ipython()
if ip:
    ip.set_hook("complete_command", completion_for_collections)



import types
import pytest

from wisefood.entities.articles import Article, ArticlesProxy

from conftest import DummyClient, StubResponse


def test_field_defaults_and_factory(dummy_client: DummyClient):
    # Include title in data to avoid lazy-fetch triggering
    article = Article(client=dummy_client, data={"urn": "urn:article:test", "title": ""}, sync=False)
    assert article.title == ""  # default value

    # default_factory values are persisted to data once accessed
    assert article.authors == []
    assert "authors" in article.data and article.data["authors"] == []


def test_read_only_field_is_protected(dummy_client: DummyClient):
    article = Article(client=dummy_client, data={"urn": "urn:article:test"}, sync=False)
    with pytest.raises(AttributeError):
        article.id = "123"


def test_get_and_create_use_normalized_urns(dummy_client: DummyClient):
    # normalize_urn returns just the slug, so endpoint is articles/abc
    dummy_client.queue_response(
        "get",
        "articles/abc",
        StubResponse(200, {"result": {"urn": "urn:article:abc", "title": "Hello"}}),
    )
    fetched = Article.get(dummy_client, "abc")
    assert fetched.urn == "urn:article:abc"
    assert fetched.data["title"] == "Hello"

    dummy_client.queue_response(
        "post",
        "articles",
        StubResponse(200, {"result": {"urn": "urn:article:new", "title": "New"}}),
    )
    created = Article.create(dummy_client, urn="new", title="New")
    assert created.urn == "urn:article:new"
    assert created.data["title"] == "New"


def test_refresh_and_delete(dummy_client: DummyClient):
    article = Article(
        client=dummy_client,
        data={"urn": "urn:article:one", "title": "Stale"},
        sync=False,
    )
    dummy_client.queue_response(
        "get",
        "articles/urn:article:one",
        StubResponse(200, {"result": {"urn": "urn:article:one", "title": "Fresh"}}),
    )
    article.refresh()
    assert article.data["title"] == "Fresh"

    article.delete()
    assert ("delete", "articles/urn:article:one", {}) in dummy_client.calls


def test_save_only_dirty_sends_changed_fields(dummy_client: DummyClient):
    article = Article(
        client=dummy_client,
        data={
            "urn": "urn:article:one",
            "id": "123",
            "creator": "alice",
            "created_at": "t1",
            "updated_at": "t1",
            "title": "Old",
            "content": "Body",
        },
        sync=False,
    )

    article.title = "New"

    dummy_client.queue_response(
        "patch",
        "articles/urn:article:one",
        StubResponse(200, {"result": {"urn": "urn:article:one", "title": "New"}}),
    )

    article.save(only_dirty=True)

    method, endpoint, json_body, _kwargs = dummy_client.calls[-1]
    assert method == "patch"
    assert endpoint == "articles/urn:article:one"
    assert json_body == {"title": "New"}
    assert article._dirty_fields == set()


def test_collection_proxy_len_iter_and_slice(monkeypatch):
    proxy = ArticlesProxy(client=None)  # type: ignore[arg-type]

    def fetch(self, limit, offset=0):
        return [f"urn:article:{i}" for i in range(offset, offset + limit)]

    def get_entity(self, urn, lazy=False):
        return f"entity:{urn}"

    monkeypatch.setattr(proxy, "_fetch_urns", types.MethodType(fetch, proxy))
    monkeypatch.setattr(proxy, "_get_entity", types.MethodType(get_entity, proxy))

    assert len(proxy) == proxy.DEFAULT_PAGE_SIZE
    assert list(proxy)[:3] == ["entity:urn:article:0", "entity:urn:article:1", "entity:urn:article:2"]

    slice_result = proxy[1:3]
    assert slice_result == ["entity:urn:article:1", "entity:urn:article:2"]


def test_collection_proxy_string_lookup_and_fuzzy(monkeypatch):
    proxy = ArticlesProxy(client=None)  # type: ignore[arg-type]
    proxy._urns = [
        "urn:article:alpha",
        "urn:article:beta",
        "urn:article:alphabet",
    ]

    def get_entity(self, urn, lazy=False):
        return f"entity:{urn}"

    monkeypatch.setattr(proxy, "_get_entity", types.MethodType(get_entity, proxy))

    # slug resolution - finds in cache, returns entity
    assert proxy["alpha"] == "entity:urn:article:alpha"
    # full URN - finds in cache, returns entity
    assert proxy["urn:article:beta"] == "entity:urn:article:beta"
    # "alph" is not in cache, so it tries to fetch from API directly
    # (fuzzy search was removed - now it fetches by slug/urn directly)
    assert proxy["alph"] == "entity:alph"


def test_slugs_exposes_known_slugs(monkeypatch):
    proxy = ArticlesProxy(client=None)  # type: ignore[arg-type]

    def fetch(self, limit, offset=0):
        return ["urn:article:first", "urn:article:second"]

    monkeypatch.setattr(proxy, "_fetch_urns", types.MethodType(fetch, proxy))
    assert proxy.slugs() == ["first", "second"]


def test_collection_proxy_create_returns_entity_and_updates_index(dummy_client: DummyClient):
    proxy = ArticlesProxy(client=dummy_client)
    proxy._urns = ["urn:article:existing"]

    dummy_client.queue_response(
        "post",
        "articles",
        StubResponse(200, {"result": {"urn": "urn:article:new", "title": "New"}}),
    )

    created = proxy.create(urn="new", title="New")

    # returns a typed entity proxy
    assert isinstance(created, Article)
    assert created.urn == "urn:article:new"

    # request was sent via the proxy's client
    method, endpoint, body, _kwargs = dummy_client.calls[-1]
    assert (method, endpoint) == ("post", "articles")
    # normalize_urn returns just the slug
    assert body["urn"] == "new"

    # cached index is refreshed
    assert "urn:article:new" in proxy._urns



import pytest

from wisefood.entities.households import (
    Household,
    HouseholdMember,
    HouseholdMemberProfile,
    HouseholdsProxy,
    MembersProxy,
)

from conftest import DummyClient, StubResponse


# ---------------------------------------------------------------------------
# HouseholdMemberProfile Tests
# ---------------------------------------------------------------------------


def test_profile_from_dict():
    data = {
        "dietary_groups": ["vegetarian", "gluten_free"],
        "nutritional_preferences": {"calories": 2000},
        "properties": {"notes": "Test"},
    }
    profile = HouseholdMemberProfile.from_dict(data)

    assert profile.dietary_groups == ["vegetarian", "gluten_free"]
    assert profile.nutritional_preferences == {"calories": 2000}
    assert profile.properties == {"notes": "Test"}


def test_profile_defaults_for_missing_fields():
    profile = HouseholdMemberProfile(data={})

    assert profile.dietary_groups == []
    assert profile.nutritional_preferences == {}
    assert profile.properties == {}


def test_profile_to_dict():
    data = {
        "dietary_groups": ["vegan"],
        "nutritional_preferences": {"protein": 50},
        "properties": {},
    }
    profile = HouseholdMemberProfile(data=data)

    assert profile.to_dict() == data


def test_profile_setattr_marks_dirty_and_syncs(dummy_client: DummyClient):
    dummy_client.queue_response(
        "patch",
        "members/member-123/profile",
        StubResponse(200, {"result": {"dietary_groups": ["vegan"]}}),
    )

    profile = HouseholdMemberProfile(
        client=dummy_client,
        member_id="member-123",
        data={"dietary_groups": []},
        sync=True,
    )

    profile.dietary_groups = ["vegan"]

    # Should have made a PATCH call
    assert len(dummy_client.calls) == 1
    method, endpoint, body, _ = dummy_client.calls[0]
    assert method == "patch"
    assert endpoint == "members/member-123/profile"
    assert body == {"dietary_groups": ["vegan"]}


def test_profile_setattr_no_sync_when_disabled(dummy_client: DummyClient):
    profile = HouseholdMemberProfile(
        client=dummy_client,
        member_id="member-123",
        data={},
        sync=False,
    )

    profile.dietary_groups = ["vegan"]

    # No API call should be made
    assert len(dummy_client.calls) == 0
    # But data should be updated locally
    assert profile.dietary_groups == ["vegan"]


def test_profile_refresh(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members/member-123/profile",
        StubResponse(200, {"result": {"dietary_groups": ["updated"]}}),
    )

    profile = HouseholdMemberProfile(
        client=dummy_client,
        member_id="member-123",
        data={"dietary_groups": ["old"]},
    )

    profile.refresh()

    assert profile.dietary_groups == ["updated"]


def test_profile_delete(dummy_client: DummyClient):
    profile = HouseholdMemberProfile(
        client=dummy_client,
        member_id="member-123",
        data={"dietary_groups": ["vegan"]},
    )

    profile.delete()

    assert ("delete", "members/member-123/profile", {}) in dummy_client.calls
    assert profile.to_dict() == {}


# ---------------------------------------------------------------------------
# HouseholdMember Tests
# ---------------------------------------------------------------------------


def test_member_from_dict(dummy_client: DummyClient):
    data = {
        "id": "member-123",
        "name": "John",
        "age_group": "adult",
        "household_id": "household-456",
        "image_url": "https://example.com/avatar.png",
        "created_at": "2024-01-01T00:00:00Z",
    }
    member = HouseholdMember.from_dict(data, dummy_client)

    assert member.id == "member-123"
    assert member.name == "John"
    assert member.age_group == "adult"
    assert member.household_id == "household-456"
    assert member.image_url == "https://example.com/avatar.png"


def test_member_property_setters_sync(dummy_client: DummyClient):
    dummy_client.queue_response(
        "patch",
        "members/member-123",
        StubResponse(200, {"result": {"id": "member-123", "name": "Jane"}}),
    )

    member = HouseholdMember(
        client=dummy_client,
        data={"id": "member-123", "name": "John", "age_group": "adult", "household_id": "h1"},
    )

    member.name = "Jane"

    method, endpoint, body, _ = dummy_client.calls[0]
    assert method == "patch"
    assert endpoint == "members/member-123"
    assert body == {"name": "Jane"}


def test_member_profile_property_fetches_on_access(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members/member-123/profile",
        StubResponse(200, {"result": {"dietary_groups": ["vegetarian"]}}),
    )

    member = HouseholdMember(
        client=dummy_client,
        data={"id": "member-123", "name": "John", "age_group": "adult", "household_id": "h1"},
    )

    profile = member.profile

    assert profile.dietary_groups == ["vegetarian"]
    # Second access should not make another API call
    _ = member.profile
    assert len(dummy_client.calls) == 1


def test_member_profile_auto_syncs_on_change(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members/member-123/profile",
        StubResponse(200, {"result": {"dietary_groups": []}}),
    )
    dummy_client.queue_response(
        "patch",
        "members/member-123/profile",
        StubResponse(200, {"result": {"dietary_groups": ["vegan"]}}),
    )

    member = HouseholdMember(
        client=dummy_client,
        data={"id": "member-123", "name": "John", "age_group": "adult", "household_id": "h1"},
    )

    member.profile.dietary_groups = ["vegan"]

    # Should have GET (fetch profile) + PATCH (update)
    assert len(dummy_client.calls) == 2
    assert dummy_client.calls[0][0] == "get"
    assert dummy_client.calls[1][0] == "patch"


def test_member_refresh(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members/member-123",
        StubResponse(200, {"result": {"id": "member-123", "name": "Updated", "age_group": "adult", "household_id": "h1"}}),
    )

    member = HouseholdMember(
        client=dummy_client,
        data={"id": "member-123", "name": "Old", "age_group": "adult", "household_id": "h1"},
    )

    member.refresh()

    assert member.name == "Updated"


def test_member_delete(dummy_client: DummyClient):
    member = HouseholdMember(
        client=dummy_client,
        data={"id": "member-123", "name": "John", "age_group": "adult", "household_id": "h1"},
    )

    member.delete()

    assert ("delete", "members/member-123", {}) in dummy_client.calls


# ---------------------------------------------------------------------------
# Household Tests
# ---------------------------------------------------------------------------


def test_household_from_dict(dummy_client: DummyClient):
    data = {
        "id": "household-123",
        "name": "My Family",
        "owner_id": "user-456",
        "region": "US",
        "metadata": {"plan": "premium"},
    }
    household = Household.from_dict(data, dummy_client)

    assert household.id == "household-123"
    assert household.name == "My Family"
    assert household.owner_id == "user-456"
    assert household.region == "US"
    assert household.metadata == {"plan": "premium"}


def test_household_property_setters_sync(dummy_client: DummyClient):
    dummy_client.queue_response(
        "patch",
        "households/household-123",
        StubResponse(200, {"result": {"id": "household-123", "name": "New Name"}}),
    )

    household = Household(
        client=dummy_client,
        data={"id": "household-123", "name": "Old Name", "owner_id": "user-1"},
    )

    household.name = "New Name"

    method, endpoint, body, _ = dummy_client.calls[0]
    assert method == "patch"
    assert endpoint == "households/household-123"
    assert body == {"name": "New Name"}


def test_household_members_property(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members",
        StubResponse(200, {"result": [
            {"id": "m1", "name": "Alice", "age_group": "adult", "household_id": "household-123"},
            {"id": "m2", "name": "Bob", "age_group": "child", "household_id": "household-123"},
        ]}),
    )

    household = Household(
        client=dummy_client,
        data={"id": "household-123", "name": "Family", "owner_id": "user-1"},
    )

    members = household.members

    assert len(members) == 2
    assert members[0].name == "Alice"
    assert members[1].name == "Bob"


def test_household_add_member(dummy_client: DummyClient):
    dummy_client.queue_response(
        "post",
        "members",
        StubResponse(200, {"result": {"id": "new-member", "name": "Charlie", "age_group": "teen", "household_id": "household-123"}}),
    )

    household = Household(
        client=dummy_client,
        data={"id": "household-123", "name": "Family", "owner_id": "user-1"},
    )

    member = household.add_member("Charlie", "teen")

    assert member.id == "new-member"
    assert member.name == "Charlie"
    assert member.age_group == "teen"

    method, endpoint, body, _ = dummy_client.calls[0]
    assert method == "post"
    assert endpoint == "members"
    assert body["name"] == "Charlie"
    assert body["age_group"] == "teen"
    assert body["household_id"] == "household-123"


def test_household_refresh(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "households/household-123",
        StubResponse(200, {"result": {"id": "household-123", "name": "Updated", "owner_id": "user-1"}}),
    )

    household = Household(
        client=dummy_client,
        data={"id": "household-123", "name": "Old", "owner_id": "user-1"},
    )

    household.refresh()

    assert household.name == "Updated"


def test_household_delete(dummy_client: DummyClient):
    household = Household(
        client=dummy_client,
        data={"id": "household-123", "name": "Family", "owner_id": "user-1"},
    )

    household.delete()

    assert ("delete", "households/household-123", {}) in dummy_client.calls


# ---------------------------------------------------------------------------
# HouseholdsProxy Tests
# ---------------------------------------------------------------------------


def test_households_proxy_me(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "households/me",
        StubResponse(200, {"result": {"id": "h1", "name": "My Home", "owner_id": "user-1"}}),
    )

    proxy = HouseholdsProxy(dummy_client)
    household = proxy.me()

    assert household.id == "h1"
    assert household.name == "My Home"


def test_households_proxy_get(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "households/h123",
        StubResponse(200, {"result": {"id": "h123", "name": "Family", "owner_id": "user-1"}}),
    )

    proxy = HouseholdsProxy(dummy_client)
    household = proxy.get("h123")

    assert household.id == "h123"


def test_households_proxy_list(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "households",
        StubResponse(200, {"result": [
            {"id": "h1", "name": "Family 1", "owner_id": "u1"},
            {"id": "h2", "name": "Family 2", "owner_id": "u2"},
        ]}),
    )

    proxy = HouseholdsProxy(dummy_client)
    households = proxy.list()

    assert len(households) == 2
    assert households[0].name == "Family 1"
    assert households[1].name == "Family 2"


def test_households_proxy_create(dummy_client: DummyClient):
    dummy_client.queue_response(
        "post",
        "households",
        StubResponse(200, {"result": {"id": "new-h", "name": "New Family", "owner_id": "user-1"}}),
    )

    proxy = HouseholdsProxy(dummy_client)
    household = proxy.create(name="New Family", region="US")

    assert household.id == "new-h"
    assert household.name == "New Family"

    method, endpoint, body, _ = dummy_client.calls[0]
    assert method == "post"
    assert body["name"] == "New Family"
    assert body["region"] == "US"


def test_households_proxy_delete(dummy_client: DummyClient):
    proxy = HouseholdsProxy(dummy_client)
    proxy.delete("h123")

    assert ("delete", "households/h123", {}) in dummy_client.calls


# ---------------------------------------------------------------------------
# MembersProxy Tests
# ---------------------------------------------------------------------------


def test_members_proxy_get(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members/m123",
        StubResponse(200, {"result": {"id": "m123", "name": "John", "age_group": "adult", "household_id": "h1"}}),
    )

    proxy = MembersProxy(dummy_client)
    member = proxy.get("m123")

    assert member.id == "m123"
    assert member.name == "John"


def test_members_proxy_list(dummy_client: DummyClient):
    dummy_client.queue_response(
        "get",
        "members",
        StubResponse(200, {"result": [
            {"id": "m1", "name": "Alice", "age_group": "adult", "household_id": "h1"},
            {"id": "m2", "name": "Bob", "age_group": "child", "household_id": "h1"},
        ]}),
    )

    proxy = MembersProxy(dummy_client)
    members = proxy.list(household_id="h1")

    assert len(members) == 2
    assert members[0].name == "Alice"


def test_members_proxy_create(dummy_client: DummyClient):
    dummy_client.queue_response(
        "post",
        "members",
        StubResponse(200, {"result": {"id": "new-m", "name": "Charlie", "age_group": "teen", "household_id": "h1"}}),
    )

    proxy = MembersProxy(dummy_client)
    member = proxy.create(household_id="h1", name="Charlie", age_group="teen")

    assert member.id == "new-m"
    assert member.name == "Charlie"


def test_members_proxy_delete(dummy_client: DummyClient):
    proxy = MembersProxy(dummy_client)
    proxy.delete("m123")

    assert ("delete", "members/m123", {}) in dummy_client.calls



# entities/articles.py
from typing import List, Optional

from .base import BaseEntity, BaseCollectionProxy, Field


class Article(BaseEntity):
    """
    Schema for /articles entities.

    All attributes here are thin accessors over `self.data`.
    """

    ENDPOINT = "articles"
    URN_PREFIX = "urn:article:"

    # Core metadata
    id: str = Field("id", read_only=True)
    title: str = Field("title", default="")
    description: Optional[str] = Field("description")
    status: str = Field("status", default="active")
    type: str = Field("type", default="article")

    # Links / identifiers
    url: Optional[str] = Field("url")
    license: Optional[str] = Field("license")
    external_id: Optional[str] = Field("external_id")
    doi: Optional[str] = Field("doi")
    organization_urn: Optional[str] = Field("organization_urn")

    # Content-ish fields
    abstract: Optional[str] = Field("abstract")
    category: Optional[str] = Field("category")
    content: str = Field("content", default="")
    venue: Optional[str] = Field("venue")
    publication_year: Optional[str] = Field("publication_year")

    # Authorship & tags
    authors: List[str] = Field("authors", default_factory=list)
    tags: List[str] = Field("tags", default_factory=list)
    ai_tags: List[str] = Field("ai_tags", default_factory=list)
    language: Optional[str] = Field("language")

    # Classification
    region: Optional[str] = Field("region")
    ai_category: Optional[str] = Field("ai_category")

    # Key takeaways
    key_takeaways: List[str] = Field("key_takeaways", default_factory=list)
    ai_key_takeaways: List[str] = Field("ai_key_takeaways", default_factory=list)

    # Timestamps / system fields
    creator: Optional[str] = Field("creator", read_only=True)
    created_at: Optional[str] = Field("created_at", read_only=True)
    updated_at: Optional[str] = Field("updated_at", read_only=True)
    embedded_at: Optional[str] = Field("embedded_at")

    # Extra metadata
    extras: Optional[dict] = Field("extras")


class ArticlesProxy(BaseCollectionProxy):
    ENTITY_CLS = Article
    ENDPOINT = "articles"



import pytest

from wisefood.client import DataClient, Credentials, WisefoodError

from conftest import StubResponse


def build_client(monkeypatch, stub_session):
    monkeypatch.setattr("wisefood.client.requests.Session", lambda: stub_session)
    return DataClient("https://api.example.com", Credentials("user", "pass"))


def test_authenticate_success(monkeypatch, stub_session, frozen_time):
    stub_session.post_response = StubResponse(
        200, {"result": {"token": "abc123", "expires_in": 120}}
    )
    client = build_client(monkeypatch, stub_session)

    assert client._token == "abc123"
    # safety_margin = 10% of expires_in = 12
    assert client._token_expiry_ts == pytest.approx(frozen_time + 120 - 12)

    assert stub_session.post_calls[0]["url"] == "https://api.example.com/api/v1/system/login"
    assert stub_session.post_calls[0]["json"] == {"username": "user", "password": "pass"}


def test_authenticate_fails_without_token(monkeypatch, stub_session):
    stub_session.post_response = StubResponse(200, {"result": {}})
    with pytest.raises(WisefoodError, match="missing token"):
        build_client(monkeypatch, stub_session)


def test_authenticate_http_error(monkeypatch, stub_session):
    stub_session.post_response = StubResponse(401, {"error": "bad"}, text="bad creds")
    with pytest.raises(WisefoodError, match="Authentication failed"):
        build_client(monkeypatch, stub_session)


def test_ensure_token_refreshes(monkeypatch, stub_session, frozen_time):
    stub_session.post_response = StubResponse(
        200, {"result": {"token": "initial", "expires_in": 5}}
    )
    client = build_client(monkeypatch, stub_session)

    calls = []

    def fake_authenticate():
        calls.append("auth")
        client._token = "refreshed"
        client._token_expiry_ts = frozen_time + 50

    client.authenticate = fake_authenticate  # type: ignore[method-assign]
    client._token = None
    client._token_expiry_ts = 0

    client._ensure_token()
    assert calls == ["auth"]
    assert client._token == "refreshed"


def test_request_adds_auth_and_merges_headers(monkeypatch, stub_session, frozen_time):
    stub_session.post_response = StubResponse(
        200, {"result": {"token": "auth-token", "expires_in": 120}}
    )
    client = build_client(monkeypatch, stub_session)

    called_with = {}

    def fake_raise_for_api_error(resp):
        called_with["resp"] = resp

    monkeypatch.setattr("wisefood.client.raise_for_api_error", fake_raise_for_api_error)
    stub_session.request_response = StubResponse(200, {"ok": True})

    resp = client.request(
        "POST",
        "items",
        headers={"Authorization": "override", "X-Test": "yes"},
        json={"hello": "world"},
    )

    assert resp is stub_session.request_response
    req = stub_session.request_calls[0]
    assert req["headers"]["Authorization"] == "Bearer auth-token"
    assert req["headers"]["X-Test"] == "yes"
    assert req["kwargs"]["json"] == {"hello": "world"}
    assert called_with["resp"] is stub_session.request_response


def test_request_without_auth(monkeypatch, stub_session):
    stub_session.post_response = StubResponse(
        200, {"result": {"token": "auth-token", "expires_in": 120}}
    )
    client = build_client(monkeypatch, stub_session)

    def boom():
        raise RuntimeError("should not be called")

    client._ensure_token = boom  # type: ignore[assignment]
    stub_session.request_response = StubResponse(200, {"ok": True})
    monkeypatch.setattr("wisefood.client.raise_for_api_error", lambda resp: None)

    client.request("GET", "items", auth=False, params={"a": 1})
    req = stub_session.request_calls[0]
    assert "Authorization" not in req["headers"]
    assert req["params"] == {"a": 1}


def test_request_rejects_body_for_get(monkeypatch, stub_session):
    stub_session.post_response = StubResponse(
        200, {"result": {"token": "auth-token", "expires_in": 120}}
    )
    client = build_client(monkeypatch, stub_session)

    with pytest.raises(ValueError):
        client.request("GET", "items", json={"a": 1})



# entities/articles.py
from typing import List, Optional

from .base import BaseEntity, BaseCollectionProxy, Field

class FCTable(BaseEntity):
    """
    Schema-backed entity for /fctables. Thin accessors over `self.data`.
    Adapted to represent a Food Composition Table (FCT).
    """

    ENDPOINT = "fctables"
    URN_PREFIX = "urn:fctable:"

    # Core metadata
    id: str = Field("id", read_only=True)
    title: str = Field("title", default="")
    description: Optional[str] = Field("description")
    status: str = Field("status", default="active")
    type: str = Field("type", default="food_composition_table")

    # Specific metadata from FoodCompositionTableSchema
    compiling_institution: str = Field("compiling_institution", default="")
    database_name: str = Field("database_name", default="")

    classification_schemes: List[str] = Field("classification_schemes", default_factory=list)
    standardization_schemes: List[str] = Field("standardization_schemes", default_factory=list)

    measurement_units: List[str] = Field("measurement_units", default_factory=list)
    reference_portions: List[str] = Field("reference_portions", default_factory=list)

    completeness_percent: Optional[float] = Field("completeness_percent")
    completeness_description: Optional[str] = Field("completeness_description")
    nutrient_coverage: List[str] = Field("nutrient_coverage", default_factory=list)

    data_formats: List[str] = Field("data_formats", default_factory=list)
    tasks_supported: List[str] = Field("tasks_supported", default_factory=list)

    number_of_entries: Optional[int] = Field("number_of_entries")
    min_nutrients_per_item: Optional[int] = Field("min_nutrients_per_item")
    max_nutrients_per_item: Optional[int] = Field("max_nutrients_per_item")

    # Links / identifiers
    url: Optional[str] = Field("url")
    license: Optional[str] = Field("license")
    external_id: Optional[str] = Field("external_id")
    organization_urn: Optional[str] = Field("organization_urn")

    # Content-ish fields (kept for compatibility)
    abstract: Optional[str] = Field("abstract")
    category: Optional[str] = Field("category")
    content: str = Field("content", default="")
    venue: str = Field("venue", default="")

    # Authorship & tags
    authors: List[str] = Field("authors", default_factory=list)
    tags: List[str] = Field("tags", default_factory=list)
    language: Optional[str] = Field("language")

    # Artifacts & region
    artifacts: List[dict] = Field("artifacts", default_factory=list)
    region: Optional[str] = Field("region")

    # Timestamps / system fields
    creator: Optional[str] = Field("creator", read_only=True)
    created_at: Optional[str] = Field("created_at", read_only=True)
    updated_at: Optional[str] = Field("updated_at", read_only=True)


class FCTablesProxy(BaseCollectionProxy):
    ENTITY_CLS = FCTable
    ENDPOINT = "fctables"





"""
Household and HouseholdMember entities and proxies for the WiseFood API.
"""

from typing import Any, Dict, List, Optional, Set, TYPE_CHECKING

if TYPE_CHECKING:
    from ..api_client import Client


# ---------------------------------------------------------------------------
# Data Models
# ---------------------------------------------------------------------------


class HouseholdMemberProfile:
    """
    Profile information for a household member.

    Setting properties automatically syncs to the API when sync=True.

    Example:
        >>> member.profile.dietary_groups = ["vegetarian", "gluten_free"]
        >>> member.profile.nutritional_preferences = {"calories": 2000}
    """

    def __init__(
        self,
        client: Optional["Client"] = None,
        member_id: Optional[str] = None,
        data: Optional[Dict[str, Any]] = None,
        sync: bool = True,
    ):
        object.__setattr__(self, "_client", client)
        object.__setattr__(self, "_member_id", member_id)
        object.__setattr__(self, "_data", data or {})
        object.__setattr__(self, "_sync", sync)
        object.__setattr__(self, "_dirty_fields", set())

    @classmethod
    def from_dict(
        cls,
        data: Dict[str, Any],
        client: Optional["Client"] = None,
        member_id: Optional[str] = None,
    ) -> "HouseholdMemberProfile":
        return cls(client=client, member_id=member_id, data=data)

    def __getattr__(self, name: str) -> Any:
        if name.startswith("_"):
            return object.__getattribute__(self, name)
        data = object.__getattribute__(self, "_data")
        if name in data:
            return data[name]
        # Return empty defaults for known fields
        if name in ("dietary_groups", "allergies"):
            return []
        if name in ("nutritional_preferences", "properties"):
            return {}
        raise AttributeError(f"'{type(self).__name__}' has no attribute '{name}'")

    def __setattr__(self, name: str, value: Any) -> None:
        if name.startswith("_"):
            object.__setattr__(self, name, value)
            return

        data = object.__getattribute__(self, "_data")
        dirty = object.__getattribute__(self, "_dirty_fields")
        sync = object.__getattribute__(self, "_sync")

        data[name] = value
        dirty.add(name)

        if sync:
            self.save()

    def save(self) -> None:
        """Persist changes to the API."""
        client = object.__getattribute__(self, "_client")
        member_id = object.__getattribute__(self, "_member_id")
        data = object.__getattribute__(self, "_data")
        dirty = object.__getattribute__(self, "_dirty_fields")

        if not client or not member_id:
            return

        if not dirty:
            return

        # Send only dirty fields
        payload = {k: data[k] for k in dirty if k in data}
        client.patch(f"members/{member_id}/profile", json=payload)
        dirty.clear()

    def refresh(self) -> None:
        """Reload profile data from the API."""
        client = object.__getattribute__(self, "_client")
        member_id = object.__getattribute__(self, "_member_id")

        if not client or not member_id:
            raise RuntimeError("Profile not bound to a client")

        resp = client.get(f"members/{member_id}/profile")
        data = resp.json().get("result", resp.json())
        object.__setattr__(self, "_data", data)
        object.__getattribute__(self, "_dirty_fields").clear()

    def delete(self) -> None:
        """Delete this profile from the API."""
        client = object.__getattribute__(self, "_client")
        member_id = object.__getattribute__(self, "_member_id")

        if not client or not member_id:
            raise RuntimeError("Profile not bound to a client")

        client.delete(f"members/{member_id}/profile")
        object.__setattr__(self, "_data", {})

    def to_dict(self) -> Dict[str, Any]:
        return dict(object.__getattribute__(self, "_data"))

    def __repr__(self) -> str:
        data = object.__getattribute__(self, "_data")
        return f"HouseholdMemberProfile({data})"


class HouseholdMember:
    """
    A member of a household.

    Access the profile via the `profile` property which auto-fetches if needed.

    Example:
        >>> member = client.members.get("member-id")
        >>> member.profile.dietary_groups = ["vegan"]
    """

    def __init__(
        self,
        client: Optional["Client"] = None,
        data: Optional[Dict[str, Any]] = None,
    ):
        self._client = client
        self._data = data or {}
        self._profile: Optional[HouseholdMemberProfile] = None

    @classmethod
    def from_dict(cls, data: Dict[str, Any], client: Optional["Client"] = None) -> "HouseholdMember":
        return cls(client=client, data=data)

    @property
    def id(self) -> str:
        return self._data["id"]

    @property
    def name(self) -> str:
        return self._data.get("name", "")

    @name.setter
    def name(self, value: str) -> None:
        self._data["name"] = value
        self._sync_field("name", value)

    @property
    def age_group(self) -> str:
        return self._data.get("age_group", "")

    @age_group.setter
    def age_group(self, value: str) -> None:
        self._data["age_group"] = value
        self._sync_field("age_group", value)

    @property
    def household_id(self) -> str:
        return self._data.get("household_id", "")

    @property
    def image_url(self) -> Optional[str]:
        return self._data.get("image_url")

    @image_url.setter
    def image_url(self, value: Optional[str]) -> None:
        self._data["image_url"] = value
        self._sync_field("image_url", value)

    @property
    def created_at(self) -> Optional[str]:
        return self._data.get("created_at")

    @property
    def updated_at(self) -> Optional[str]:
        return self._data.get("updated_at")

    @property
    def profile(self) -> HouseholdMemberProfile:
        """
        Get the member's profile, fetching from API if not loaded.

        Returns a profile object where setting properties auto-syncs to API.

        Example:
            >>> member.profile.dietary_groups = ["vegetarian"]
            >>> member.profile.nutritional_preferences = {"protein": 50}
        """
        if self._profile is None:
            if not self._client:
                raise RuntimeError("Member not bound to a client")
            try:
                resp = self._client.get(f"members/{self.id}/profile")
                data = resp.json().get("result", resp.json())
            except Exception:
                # Profile might not exist yet, return empty profile
                data = {}
            self._profile = HouseholdMemberProfile(
                client=self._client,
                member_id=self.id,
                data=data,
            )
        return self._profile

    def _sync_field(self, field: str, value: Any) -> None:
        """Sync a single field to the API."""
        if self._client:
            self._client.patch(f"members/{self.id}", json={field: value})

    def refresh(self) -> None:
        """Reload member data from the API."""
        if not self._client:
            raise RuntimeError("Member not bound to a client")
        resp = self._client.get(f"members/{self.id}")
        self._data = resp.json().get("result", resp.json())
        self._profile = None  # Reset profile to force re-fetch

    def delete(self) -> None:
        """Delete this member."""
        if not self._client:
            raise RuntimeError("Member not bound to a client")
        self._client.delete(f"members/{self.id}")

    def to_dict(self) -> Dict[str, Any]:
        return dict(self._data)

    def __repr__(self) -> str:
        return f"<HouseholdMember id='{self.id}' name='{self.name}'>"


class Household:
    """
    A household in the WiseFood system.

    Example:
        >>> household = client.households.me()
        >>> household.name = "New Family Name"  # auto-syncs
        >>> members = household.members
    """

    def __init__(
        self,
        client: Optional["Client"] = None,
        data: Optional[Dict[str, Any]] = None,
    ):
        self._client = client
        self._data = data or {}

    @classmethod
    def from_dict(cls, data: Dict[str, Any], client: Optional["Client"] = None) -> "Household":
        return cls(client=client, data=data)

    @property
    def id(self) -> str:
        return self._data["id"]

    @property
    def name(self) -> str:
        return self._data.get("name", "")

    @name.setter
    def name(self, value: str) -> None:
        self._data["name"] = value
        self._sync_field("name", value)

    @property
    def owner_id(self) -> str:
        return self._data.get("owner_id", "")

    @property
    def region(self) -> Optional[str]:
        return self._data.get("region")

    @region.setter
    def region(self, value: Optional[str]) -> None:
        self._data["region"] = value
        self._sync_field("region", value)

    @property
    def metadata(self) -> Optional[Dict[str, Any]]:
        return self._data.get("metadata")

    @metadata.setter
    def metadata(self, value: Optional[Dict[str, Any]]) -> None:
        self._data["metadata"] = value
        self._sync_field("metadata", value)

    @property
    def created_at(self) -> Optional[str]:
        return self._data.get("created_at")

    @property
    def updated_at(self) -> Optional[str]:
        return self._data.get("updated_at")

    @property
    def members(self) -> List["HouseholdMember"]:
        """Get all members of this household."""
        if not self._client:
            raise RuntimeError("Household not bound to a client")
        resp = self._client.get("members", household_id=self.id)
        data = resp.json().get("result", resp.json())
        if isinstance(data, list):
            return [HouseholdMember.from_dict(m, self._client) for m in data]
        return []

    def _sync_field(self, field: str, value: Any) -> None:
        """Sync a single field to the API."""
        if self._client:
            self._client.patch(f"households/{self.id}", json={field: value})

    def add_member(self, name: str, age_group: str, **kwargs) -> "HouseholdMember":
        """Add a new member to this household."""
        if not self._client:
            raise RuntimeError("Household not bound to a client")
        payload = {"name": name, "age_group": age_group, "household_id": self.id, **kwargs}
        resp = self._client.post("members", json=payload)
        data = resp.json().get("result", resp.json())
        return HouseholdMember.from_dict(data, self._client)

    def refresh(self) -> None:
        """Reload household data from the API."""
        if not self._client:
            raise RuntimeError("Household not bound to a client")
        resp = self._client.get(f"households/{self.id}")
        self._data = resp.json().get("result", resp.json())

    def delete(self) -> None:
        """Delete this household."""
        if not self._client:
            raise RuntimeError("Household not bound to a client")
        self._client.delete(f"households/{self.id}")

    def to_dict(self) -> Dict[str, Any]:
        return dict(self._data)

    def __repr__(self) -> str:
        return f"<Household id='{self.id}' name='{self.name}'>"


# ---------------------------------------------------------------------------
# Proxy Classes
# ---------------------------------------------------------------------------


class HouseholdsProxy:
    """
    Proxy for household operations.

    Provides convenient access to household CRUD operations through the Client.

    Example:
        >>> client.households.me()  # Get current user's household
        >>> client.households.get("household-id-123")
        >>> client.households.create(name="My Family")
    """

    def __init__(self, client: "Client") -> None:
        self._client = client

    def me(self) -> Household:
        """
        Get the household owned by the authenticated user.

        Returns:
            Household object for the current user

        Example:
            >>> household = client.households.me()
            >>> print(household.name)
        """
        resp = self._client.get("households/me")
        data = resp.json().get("result", resp.json())
        return Household.from_dict(data, self._client)

    def get(self, household_id: str) -> Household:
        """
        Get a household by ID.

        Args:
            household_id: The household's unique identifier

        Returns:
            Household object

        Example:
            >>> household = client.households.get("abc123")
        """
        resp = self._client.get(f"households/{household_id}")
        data = resp.json().get("result", resp.json())
        return Household.from_dict(data, self._client)

    def list(self, limit: int = 100, offset: int = 0) -> List[Household]:
        """
        List all households (admin only).

        Args:
            limit: Maximum number of results (default: 100)
            offset: Number of results to skip (default: 0)

        Returns:
            List of Household objects

        Example:
            >>> households = client.households.list(limit=50)
        """
        resp = self._client.get("households", limit=limit, offset=offset)
        data = resp.json().get("result", resp.json())
        if isinstance(data, list):
            return [Household.from_dict(h, self._client) for h in data]
        return []

    def create(self, name: str, region: Optional[str] = None,
               metadata: Optional[Dict[str, Any]] = None,
               members: Optional[List[Dict[str, Any]]] = None) -> Household:
        """
        Create a new household.

        Args:
            name: Household name
            region: Optional region identifier
            metadata: Optional metadata dictionary
            members: Optional list of initial members to create

        Returns:
            Created Household object

        Example:
            >>> household = client.households.create(
            ...     name="My Family",
            ...     region="US-CA",
            ...     members=[{"name": "John", "age_group": "adult"}]
            ... )
        """
        payload: Dict[str, Any] = {"name": name}
        if region:
            payload["region"] = region
        if metadata:
            payload["metadata"] = metadata
        if members:
            payload["members"] = members
        resp = self._client.post("households", json=payload)
        data = resp.json().get("result", resp.json())
        return Household.from_dict(data, self._client)

    def update(self, household_id: str, **kwargs) -> Household:
        """
        Update a household.

        Args:
            household_id: The household's unique identifier
            **kwargs: Fields to update (name, region, metadata)

        Returns:
            Updated Household object

        Example:
            >>> household = client.households.update("abc123", name="New Name")
        """
        resp = self._client.patch(f"households/{household_id}", json=kwargs)
        data = resp.json().get("result", resp.json())
        return Household.from_dict(data, self._client)

    def delete(self, household_id: str) -> None:
        """
        Delete a household.

        Args:
            household_id: The household's unique identifier

        Example:
            >>> client.households.delete("abc123")
        """
        self._client.delete(f"households/{household_id}")


class MembersProxy:
    """
    Proxy for household member operations.

    Provides convenient access to member CRUD operations through the Client.

    Example:
        >>> member = client.members.get("member-id-123")
        >>> member.profile.dietary_groups = ["vegetarian"]  # auto-syncs
    """

    def __init__(self, client: "Client") -> None:
        self._client = client

    def get(self, member_id: str) -> HouseholdMember:
        """
        Get a household member by ID.

        Args:
            member_id: The member's unique identifier

        Returns:
            HouseholdMember object with auto-syncing profile

        Example:
            >>> member = client.members.get("member123")
            >>> member.profile.dietary_groups = ["vegan"]
        """
        resp = self._client.get(f"members/{member_id}")
        data = resp.json().get("result", resp.json())
        return HouseholdMember.from_dict(data, self._client)

    def list(self, household_id: str, limit: int = 100, offset: int = 0) -> List[HouseholdMember]:
        """
        List household members.

        Args:
            household_id: The household's unique identifier
            limit: Maximum number of results (default: 100)
            offset: Number of results to skip (default: 0)

        Returns:
            List of HouseholdMember objects

        Example:
            >>> members = client.members.list(household_id="abc123")
        """
        resp = self._client.get("members", household_id=household_id, limit=limit, offset=offset)
        data = resp.json().get("result", resp.json())
        if isinstance(data, list):
            return [HouseholdMember.from_dict(m, self._client) for m in data]
        return []

    def create(self, household_id: str, name: str, age_group: str,
               image_url: Optional[str] = None) -> HouseholdMember:
        """
        Create a new household member.

        Args:
            household_id: The household's unique identifier
            name: Member name
            age_group: Age group (child, teen, adult, senior, etc.)
            image_url: Optional URL to member's profile image

        Returns:
            Created HouseholdMember object

        Example:
            >>> member = client.members.create(
            ...     household_id="abc123",
            ...     name="John",
            ...     age_group="adult"
            ... )
            >>> member.profile.dietary_groups = ["vegetarian"]
        """
        payload: Dict[str, Any] = {
            "household_id": household_id,
            "name": name,
            "age_group": age_group,
        }
        if image_url:
            payload["image_url"] = image_url
        resp = self._client.post("members", json=payload)
        data = resp.json().get("result", resp.json())
        return HouseholdMember.from_dict(data, self._client)

    def delete(self, member_id: str) -> None:
        """
        Delete a household member.

        Args:
            member_id: The member's unique identifier

        Example:
            >>> client.members.delete("member123")
        """
        self._client.delete(f"members/{member_id}")



"""
A compact, robust HTTP client for the Wisefood API (systemic operations).
"""

from dataclasses import dataclass
import time
from typing import Any, Dict, Optional
import urllib.parse
import requests
from requests.adapters import HTTPAdapter
from urllib3.util.retry import Retry
from .exceptions import raise_for_api_error
from .entities.households import HouseholdsProxy, MembersProxy


# -------------------------------
# Credentials Model
# -------------------------------



@dataclass
class Credentials:
    """
    Either user credentials (username & password) OR client credentials
    (client_id & client_secret) must be provided. They are mutually exclusive.
    """
    username: Optional[str] = None
    password: Optional[str] = None
    client_id: Optional[str] = None
    client_secret: Optional[str] = None

    def __post_init__(self) -> None:
        has_user = bool(self.username or self.password)
        has_client = bool(self.client_id or self.client_secret)

        if has_user and has_client:
            raise ValueError("Provide either username/password OR client_id/client_secret, not both.")

        if not (self.username and self.password) and not (self.client_id and self.client_secret):
            raise ValueError("Must provide either username/password OR client_id/client_secret.")

    @property
    def is_user_credentials(self) -> bool:
        return bool(self.username and self.password)

    @property
    def is_client_credentials(self) -> bool:
        return bool(self.client_id and self.client_secret)


# -------------------------------
# Exceptions
# -------------------------------


class WisefoodError(RuntimeError):
    pass


# -------------------------------
# Main Client
# -------------------------------


class Client:
    """
    HTTP client for the Wisefood API with automatic authentication and token management.

    This client handles all communication with the Wisefood API, including:
    - Automatic authentication using username/password or client credentials
    - Token refresh when expired
    - Connection pooling and retry logic
    - Clean endpoint URL construction

    Args:
        base_url: Base URL of the Wisefood API (e.g., 'https://api.wisefood.com')
        credentials: Credentials object containing either username/password or client_id/client_secret
        api_prefix: API version prefix (default: '/api/v1')
        verify_tls: Whether to verify SSL/TLS certificates (default: True)
        default_timeout: Default request timeout in seconds (default: 30.0)
        pool_connections: Number of connection pools to cache (default: 3)
        pool_maxsize: Maximum number of connections to save in the pool (default: 10)

    Example:
        >>> creds = Credentials(username='user@example.com', password='secret')
        >>> client = Client('https://api.wisefood.com', creds)
        >>> response = client.GET('foods', 'search', q='apple')
        >>> foods = response.json()
    """

    def __init__(
        self,
        base_url: str,
        credentials: Credentials,
        *,
        api_prefix: str = "/api/v1",
        verify_tls: bool = True,
        default_timeout: float = 30.0,
        pool_connections: int = 3,
        pool_maxsize: int = 10,
    ) -> None:

        self.base_url = base_url.rstrip("/")
        self.api_prefix = api_prefix.strip("/")
        self.credentials = credentials
        self.verify_tls = verify_tls
        self.default_timeout = default_timeout

        self._session = requests.Session()

        # Configure connection pooling
        retry_strategy = Retry(
            total=3,
            backoff_factor=1,
            status_forcelist=[429, 500, 502, 503, 504],
        )
        adapter = HTTPAdapter(
            pool_connections=pool_connections,
            pool_maxsize=pool_maxsize,
            max_retries=retry_strategy,
        )
        self._session.mount("http://", adapter)
        self._session.mount("https://", adapter)

        self._token: Optional[str] = None
        self._token_expiry_ts: float = 0.0

        # Authenticate immediately
        self.authenticate()

        # Proxies for API resource groups
        self.households = HouseholdsProxy(self)
        self.members = MembersProxy(self)

    # ------------------------------------------------------------------
    # URL helpers
    # ------------------------------------------------------------------

    def _join(self, base: str, path: str) -> str:
        """
        Join base and path cleanly without stripping segments.

        Args:
            base: Base URL or path
            path: Path to append

        Returns:
            Properly joined URL path
        """
        return urllib.parse.urljoin(base.rstrip("/") + "/", path.lstrip("/"))

    @property
    def api_base(self) -> str:
        """
        Get the full API base URL combining base_url and api_prefix.

        Returns:
            Complete API base URL (e.g., 'https://api.wisefood.com/api/v1')
        """
        return self._join(self.base_url, self.api_prefix)

    def endpoint(self, endpoint: str) -> str:
        """
        Construct absolute URL for an API endpoint.

        Args:
            endpoint: Relative endpoint path (e.g., 'foods/123' or '/foods/123')

        Returns:
            Complete URL (e.g., 'https://api.wisefood.com/api/v1/foods/123')
        """
        return self._join(self.api_base, endpoint)

    # ------------------------------------------------------------------
    # Authentication
    # ------------------------------------------------------------------

    def authenticate(self) -> None:
        """
        Authenticate with the API and store bearer token with expiry timestamp.

        Uses either username/password (user credentials) or client_id/client_secret
        (machine-to-machine credentials) depending on the credentials type.

        The token is stored internally with an automatic safety margin before expiry
        to ensure requests don't fail due to token expiration.

        Raises:
            WisefoodError: If authentication fails or response is invalid
        """

        if self.credentials.is_client_credentials:
            url = self.endpoint("system/mtm")
            payload = {
                "client_id": self.credentials.client_id,
                "client_secret": self.credentials.client_secret,
            }
        else:
            url = self.endpoint("system/login")
            payload = {
                "username": self.credentials.username,
                "password": self.credentials.password,
            }

        resp = self._session.post(
            url,
            json=payload,
            verify=self.verify_tls,
            timeout=self.default_timeout,
        )

        if resp.status_code != 200:
            raise WisefoodError(
                f"Authentication failed ({resp.status_code}): {resp.text}"
            )

        data = resp.json().get("result", {})
        token = data.get("token") or data.get("access_token") or data.get("jwt")

        if not token:
            raise WisefoodError("Authentication response missing token field.")

        expires_in = float(data.get("expires_in", 3600))
        now = time.time()
        safety_margin = min(60, max(10, expires_in * 0.1))

        self._token = token
        self._token_expiry_ts = now + expires_in - safety_margin

    def _ensure_token(self) -> None:
        """
        Ensure a valid authentication token exists, refreshing if necessary.

        Automatically re-authenticates if the token is missing or expired.
        """
        if not self._token or time.time() >= self._token_expiry_ts:
            self.authenticate()

    def ping(self) -> Dict[str, Any]:
        """
        Check authentication status and get user/client information.

        Returns:
            Dictionary containing authentication status and user/client details

        Example:
            >>> status = client.ping()
            >>> print(status.get('username'))
        """
        return self.GET("system/ping").json().get("result", {})

    # ------------------------------------------------------------------
    # Low-level request
    # ------------------------------------------------------------------

    def request(
        self,
        method: str,
        endpoint: str,
        *,
        auth=True,
        timeout=None,
        headers=None,
        params=None,
        **kwargs,
    ):
        """
        Low-level HTTP request method with automatic authentication.

        Args:
            method: HTTP method (GET, POST, PUT, PATCH, DELETE)
            endpoint: API endpoint path relative to api_base
            auth: Whether to include authentication token (default: True)
            timeout: Request timeout in seconds (uses default_timeout if None)
            headers: Additional HTTP headers to include
            params: Query parameters for the request
            **kwargs: Additional arguments passed to requests (e.g., json, data)

        Returns:
            requests.Response object

        Raises:
            ValueError: If GET/DELETE request includes a body
            WisefoodError: If the API returns an error response

        Example:
            >>> response = client.request('GET', 'foods/123')
            >>> response = client.request('POST', 'foods', json={'name': 'Apple'})
        """
        url = self.endpoint(endpoint)

        req_headers: Dict[str, str] = {}
        if auth:
            self._ensure_token()
            req_headers["Authorization"] = f"Bearer {self._token}"

        # Merge headers but avoid overriding Authorization
        if headers:
            filtered = {
                k: v
                for k, v in headers.items()
                if not (auth and k.lower() == "authorization")
            }
            req_headers.update(filtered)

        # GET/DELETE must not have bodies
        if kwargs and method.upper() in {"GET", "DELETE"}:
            raise ValueError("GET and DELETE requests cannot include a request body.")

        resp = self._session.request(
            method.upper(),
            url,
            headers=req_headers,
            params=params,
            verify=self.verify_tls,
            timeout=self.default_timeout if timeout is None else timeout,
            **kwargs,
        )
        raise_for_api_error(resp)

        return resp

    # ------------------------------------------------------------------
    # API-relative HTTP verbs
    # ------------------------------------------------------------------

    def get(self, endpoint: str, **params: Any) -> requests.Response:
        """
        Perform a GET request to the specified endpoint.

        Args:
            endpoint: API endpoint path
            **params: Query parameters as keyword arguments

        Returns:
            requests.Response object

        Example:
            >>> response = client.get('foods/search', q='apple', limit=10)
        """
        return self.request("GET", endpoint, params=params)

    def post(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        Perform a POST request to the specified endpoint.

        Args:
            endpoint: API endpoint path
            **kwargs: Request arguments (typically json=dict or data=dict)

        Returns:
            requests.Response object

        Example:
            >>> response = client.post('foods', json={'name': 'Apple', 'calories': 95})
        """
        return self.request("POST", endpoint, **kwargs)

    def put(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        Perform a PUT request to the specified endpoint.

        Args:
            endpoint: API endpoint path
            **kwargs: Request arguments (typically json=dict or data=dict)

        Returns:
            requests.Response object

        Example:
            >>> response = client.put('foods/123', json={'name': 'Green Apple'})
        """
        return self.request("PUT", endpoint, **kwargs)

    def patch(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        Perform a PATCH request to the specified endpoint.

        Args:
            endpoint: API endpoint path
            **kwargs: Request arguments (typically json=dict or data=dict)

        Returns:
            requests.Response object

        Example:
            >>> response = client.patch('foods/123', json={'calories': 100})
        """
        return self.request("PATCH", endpoint, **kwargs)

    def delete(self, endpoint: str, **kwargs: Any) -> requests.Response:
        """
        Perform a DELETE request to the specified endpoint.

        Args:
            endpoint: API endpoint path
            **kwargs: Request arguments (typically params for query parameters)

        Returns:
            requests.Response object

        Example:
            >>> response = client.delete('foods/123')
        """
        return self.request("DELETE", endpoint, **kwargs)

    # ------------------------------------------------------------------
    # Wrappers
    # ------------------------------------------------------------------

    def GET(self, *parts, **params) -> requests.Response:
        """
        Convenient GET request with path parts as separate arguments.

        Args:
            *parts: URL path segments that will be joined with '/'
            **params: Query parameters as keyword arguments

        Returns:
            requests.Response object

        Example:
            >>> response = client.GET('foods', 'search', q='apple', limit=10)
            # Equivalent to: GET /api/v1/foods/search?q=apple&limit=10
        """
        endpoint = "/".join(str(p) for p in parts)
        return self.get(endpoint, params=params)

    def POST(self, *parts, params=None, **json) -> requests.Response:
        """
        Convenient POST request with path parts and JSON body.

        Args:
            *parts: URL path segments that will be joined with '/'
            params: Optional query parameters
            **json: JSON body fields as keyword arguments

        Returns:
            requests.Response object

        Example:
            >>> response = client.POST('foods', name='Apple', calories=95)
            # Equivalent to: POST /api/v1/foods with JSON body
        """
        endpoint = "/".join(str(p) for p in parts)
        return self.post(endpoint, params=params, json=json)

    def PUT(self, *parts, params=None, **json) -> requests.Response:
        """
        Convenient PUT request with path parts and JSON body.

        Args:
            *parts: URL path segments that will be joined with '/'
            params: Optional query parameters
            **json: JSON body fields as keyword arguments

        Returns:
            requests.Response object

        Example:
            >>> response = client.PUT('foods', '123', name='Green Apple')
            # Equivalent to: PUT /api/v1/foods/123 with JSON body
        """
        endpoint = "/".join(str(p) for p in parts)
        return self.put(endpoint, params=params, json=json)

    def PATCH(self, *parts, params=None, **json) -> requests.Response:
        """
        Convenient PATCH request with path parts and JSON body.

        Args:
            *parts: URL path segments that will be joined with '/'
            params: Optional query parameters
            **json: JSON body fields as keyword arguments

        Returns:
            requests.Response object

        Example:
            >>> response = client.PATCH('foods', '123', calories=100)
            # Equivalent to: PATCH /api/v1/foods/123 with JSON body
        """
        endpoint = "/".join(str(p) for p in parts)
        return self.patch(endpoint, params=params, json=json)

    def DELETE(self, *parts, **params) -> requests.Response:
        """
        Convenient DELETE request with path parts as separate arguments.

        Args:
            *parts: URL path segments that will be joined with '/'
            **params: Query parameters as keyword arguments

        Returns:
            requests.Response object

        Example:
            >>> response = client.DELETE('foods', '123')
            # Equivalent to: DELETE /api/v1/foods/123
        """
        endpoint = "/".join(str(p) for p in parts)
        return self.delete(endpoint, params=params)
